from __future__ import annotations
from typing import Any
from ..arctrl_core.arc_types import ArcWorkflow
from ..arctrl_core.Helper.identifier import (Workflow_fileNameFromIdentifier, Workflow_cwlFileNameFromIdentifier)
from ..arctrl_cwl.cwlprocessing_unit import CWLProcessingUnit
from ..arctrl_cwl.decode import Decode_decodeCWLProcessingUnit
from ..arctrl_cwl.encode import encode_processing_unit
from ..arctrl_file_system.file_system_tree import FileSystemTree
from ..arctrl_file_system.path import (combine_many, get_workflow_folder_path)
from ..arctrl_spreadsheet.arc_workflow import (ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E, ARCtrl_ArcWorkflow__ArcWorkflow_fromFsWorkbook_Static_32154C9D)
from ..fable_library.array_ import equals_with
from ..fable_library.option import (default_arg, value)
from ..fable_library.seq import (to_array, delay, append, collect, singleton, empty)
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from .contract import (Contract, DTOType, DTO)

def _007CWorkflowPath_007C__007C(input: Array[str]) -> str | None:
    (pattern_matching_result,) = (None,)
    def _arrow3733(x: str, y: str, input: Any=input) -> bool:
        return x == y

    if (len(input) == 3) if (not equals_with(_arrow3733, input, None)) else False:
        if input[0] == "workflows":
            if input[2] == "isa.workflow.xlsx":
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        any_workflow_name: str = input[1]
        return combine_many(input)

    elif pattern_matching_result == 1:
        return None



def _007CWorkflowCWLPath_007C__007C(input: Array[str]) -> str | None:
    (pattern_matching_result,) = (None,)
    def _arrow3734(x: str, y: str, input: Any=input) -> bool:
        return x == y

    if (len(input) == 3) if (not equals_with(_arrow3734, input, None)) else False:
        if input[0] == "workflows":
            if input[2] == "workflow.cwl":
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        any_workflow_name: str = input[1]
        return combine_many(input)

    elif pattern_matching_result == 1:
        return None



def ARCtrl_ArcWorkflow__ArcWorkflow_ToCreateContract_6FCE9E49(this: ArcWorkflow, WithFolder: bool | None=None) -> Array[Contract]:
    with_folder: bool = default_arg(WithFolder, False)
    path: str = Workflow_fileNameFromIdentifier(this.Identifier)
    c: Contract = Contract.create_create(path, DTOType(2), DTO(0, ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E(this)))
    def _arrow3740(__unit: None=None, this: Any=this, WithFolder: Any=WithFolder) -> IEnumerable_1[Contract]:
        def _arrow3736(__unit: None=None) -> IEnumerable_1[Contract]:
            folder_fs: FileSystemTree = FileSystemTree.create_workflows_folder([FileSystemTree.create_workflow_folder(this.Identifier)])
            def _arrow3735(p: str) -> IEnumerable_1[Contract]:
                return singleton(Contract.create_create(p, DTOType(10))) if ((p != "workflows/.gitkeep") if (p != path) else False) else empty()

            return collect(_arrow3735, folder_fs.ToFilePaths(False))

        def _arrow3739(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow3737(__unit: None=None) -> IEnumerable_1[Contract]:
                path_1: str = Workflow_cwlFileNameFromIdentifier(this.Identifier)
                dto: str = encode_processing_unit(value(this.CWLDescription))
                return singleton(Contract.create_create(path_1, DTOType(8), DTO(1, dto)))

            def _arrow3738(__unit: None=None) -> IEnumerable_1[Contract]:
                return singleton(c)

            return append(_arrow3737() if (this.CWLDescription is not None) else empty(), delay(_arrow3738))

        return append(_arrow3736() if with_folder else empty(), delay(_arrow3739))

    return to_array(delay(_arrow3740))


def ARCtrl_ArcWorkflow__ArcWorkflow_ToUpdateContract(this: ArcWorkflow) -> Contract:
    path: str = Workflow_fileNameFromIdentifier(this.Identifier)
    return Contract.create_update(path, DTOType(2), DTO(0, ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E(this)))


def ARCtrl_ArcWorkflow__ArcWorkflow_ToDeleteContract(this: ArcWorkflow) -> Contract:
    path: str = get_workflow_folder_path(this.Identifier)
    return Contract.create_delete(path)


def ARCtrl_ArcWorkflow__ArcWorkflow_toDeleteContract_Static_Z1C75CB0E(workflow: ArcWorkflow) -> Contract:
    return ARCtrl_ArcWorkflow__ArcWorkflow_ToDeleteContract(workflow)


def ARCtrl_ArcWorkflow__ArcWorkflow_toCreateContract_Static_3B1E4D7B(workflow: ArcWorkflow, WithFolder: bool | None=None) -> Array[Contract]:
    return ARCtrl_ArcWorkflow__ArcWorkflow_ToCreateContract_6FCE9E49(workflow, WithFolder)


def ARCtrl_ArcWorkflow__ArcWorkflow_toUpdateContract_Static_Z1C75CB0E(workflow: ArcWorkflow) -> Contract:
    return ARCtrl_ArcWorkflow__ArcWorkflow_ToUpdateContract(workflow)


def ARCtrl_ArcWorkflow__ArcWorkflow_tryFromReadContract_Static_7570923F(c: Contract) -> ArcWorkflow | None:
    (pattern_matching_result, fsworkbook) = (None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 2:
                if c.DTO is not None:
                    if c.DTO.tag == 0:
                        pattern_matching_result = 0
                        fsworkbook = c.DTO.fields[0]

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return ARCtrl_ArcWorkflow__ArcWorkflow_fromFsWorkbook_Static_32154C9D(fsworkbook)

    elif pattern_matching_result == 1:
        return None



def ARCtrl_ArcWorkflow__ArcWorkflow_tryCWLFromReadContract_Static(workflow_identifier: str, c: Contract) -> CWLProcessingUnit | None:
    p: str = Workflow_cwlFileNameFromIdentifier(workflow_identifier)
    (pattern_matching_result, cwl_1) = (None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 8:
                if c.DTO is not None:
                    if c.DTO.tag == 1:
                        def _arrow3741(__unit: None=None, workflow_identifier: Any=workflow_identifier, c: Any=c) -> bool:
                            cwl: str = c.DTO.fields[0]
                            return c.Path == p

                        if _arrow3741():
                            pattern_matching_result = 0
                            cwl_1 = c.DTO.fields[0]

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return Decode_decodeCWLProcessingUnit(cwl_1)

    elif pattern_matching_result == 1:
        return None



__all__ = ["_007CWorkflowPath_007C__007C", "_007CWorkflowCWLPath_007C__007C", "ARCtrl_ArcWorkflow__ArcWorkflow_ToCreateContract_6FCE9E49", "ARCtrl_ArcWorkflow__ArcWorkflow_ToUpdateContract", "ARCtrl_ArcWorkflow__ArcWorkflow_ToDeleteContract", "ARCtrl_ArcWorkflow__ArcWorkflow_toDeleteContract_Static_Z1C75CB0E", "ARCtrl_ArcWorkflow__ArcWorkflow_toCreateContract_Static_3B1E4D7B", "ARCtrl_ArcWorkflow__ArcWorkflow_toUpdateContract_Static_Z1C75CB0E", "ARCtrl_ArcWorkflow__ArcWorkflow_tryFromReadContract_Static_7570923F", "ARCtrl_ArcWorkflow__ArcWorkflow_tryCWLFromReadContract_Static"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.list import (reduce, map, FSharpList, empty, append, is_empty, tail, head)
from ...fable_library.reflection import (TypeInfo, char_type, union_type, class_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (Array, Union, to_string)
from ..Cells.fs_cell import DataType
from .expression import (RequiredSource_1__ctor_2B595, RequiredSource_1, OptionalSource_1__ctor_2B595, OptionalSource_1, RequiredSource_1__get_Source, OptionalSource_1__get_Source)
from .types import (SheetEntity_1, Message__MapText_11D407F6, Message, SheetEntity_1__get_Messages)

def _expr309() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.ReduceOperation", [], ReduceOperation, lambda: [[("Item", char_type)]])


class ReduceOperation(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Concat"]


ReduceOperation_reflection = _expr309

def ReduceOperation__Reduce_70D1A69E(this: ReduceOperation, values: FSharpList[tuple[DataType, Any]]) -> tuple[DataType, Any]:
    def reduction(a: str, b: str, this: Any=this, values: Any=values) -> str:
        return ((((("" + a) + "") + str(this.fields[0])) + "") + b) + ""

    def mapping(arg: tuple[DataType, Any], this: Any=this, values: Any=values) -> str:
        return to_string(arg[1])

    return (DataType(0), reduce(reduction, map(mapping, values)))


def _expr310() -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.CellBuilder", None, CellBuilder)


class CellBuilder:
    def __init__(self, __unit: None=None) -> None:
        self.reducer: ReduceOperation = ReduceOperation(0, ",")


CellBuilder_reflection = _expr310

def CellBuilder__ctor(__unit: None=None) -> CellBuilder:
    return CellBuilder(__unit)


def CellBuilder_get_Empty(__unit: None=None) -> SheetEntity_1[FSharpList[tuple[DataType, Any]]]:
    return SheetEntity_1(1, empty())


def CellBuilder__SignMessages_3F89FFFC(this: CellBuilder, messages: FSharpList[Message]) -> FSharpList[Message]:
    def mapping(m: Message, this: Any=this, messages: Any=messages) -> Message:
        def _arrow311(__unit: None=None, m: Any=m) -> Callable[[str], str]:
            clo: Callable[[str], str] = to_text(printf("In Cell: %s"))
            return clo

        return Message__MapText_11D407F6(m, _arrow311())

    return map(mapping, messages)


def CellBuilder__Yield_36B9E420(this: CellBuilder, ro: ReduceOperation) -> SheetEntity_1[FSharpList[tuple[DataType, Any]]]:
    this.reducer = ro
    return SheetEntity_1(1, empty())


def CellBuilder__Combine_F07E260(this: CellBuilder, wx1: SheetEntity_1[FSharpList[tuple[DataType, Any]]], wx2: SheetEntity_1[FSharpList[tuple[DataType, Any]]]) -> SheetEntity_1[FSharpList[tuple[DataType, Any]]]:
    (pattern_matching_result, l1, l2, messages1, messages2, messages2_1, messages1_1, f1, messages1_2, messages2_2, f2, messages1_3, messages2_3, messages1_4, messages2_4) = (None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)
    if wx1.tag == 2:
        if wx2.tag == 2:
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]

        else: 
            pattern_matching_result = 2
            messages1_1 = wx1.fields[0]


    elif wx1.tag == 1:
        if wx2.tag == 0:
            pattern_matching_result = 4
            f2 = wx2.fields[0]
            messages1_3 = wx1.fields[0]
            messages2_3 = wx2.fields[1]

        elif wx2.tag == 1:
            pattern_matching_result = 5
            messages1_4 = wx1.fields[0]
            messages2_4 = wx2.fields[0]

        else: 
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]


    elif wx2.tag == 2:
        pattern_matching_result = 1
        messages2_1 = wx2.fields[0]

    elif wx2.tag == 1:
        pattern_matching_result = 3
        f1 = wx1.fields[0]
        messages1_2 = wx1.fields[1]
        messages2_2 = wx2.fields[0]

    else: 
        pattern_matching_result = 0
        l1 = wx1.fields[0]
        l2 = wx2.fields[0]
        messages1 = wx1.fields[1]
        messages2 = wx2.fields[1]

    if pattern_matching_result == 0:
        return SheetEntity_1(0, append(l1, l2), append(messages1, messages2))

    elif pattern_matching_result == 1:
        return SheetEntity_1(2, append(SheetEntity_1__get_Messages(wx1), messages2_1))

    elif pattern_matching_result == 2:
        return SheetEntity_1(2, append(messages1_1, SheetEntity_1__get_Messages(wx2)))

    elif pattern_matching_result == 3:
        return SheetEntity_1(0, f1, append(messages1_2, messages2_2))

    elif pattern_matching_result == 4:
        return SheetEntity_1(0, f2, append(messages1_3, messages2_3))

    elif pattern_matching_result == 5:
        return SheetEntity_1(1, append(messages1_4, messages2_4))



def CellBuilder__Combine_4BCE267E(this: CellBuilder, wx1: RequiredSource_1[None], wx2: SheetEntity_1[FSharpList[tuple[DataType, Any]]]) -> RequiredSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]:
    return RequiredSource_1__ctor_2B595(wx2)


def CellBuilder__Combine_Z2CF9C142(this: CellBuilder, wx1: SheetEntity_1[FSharpList[tuple[DataType, Any]]], wx2: RequiredSource_1[None]) -> RequiredSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]:
    return RequiredSource_1__ctor_2B595(wx1)


def CellBuilder__Combine_583F6A19(this: CellBuilder, wx1: OptionalSource_1[None], wx2: SheetEntity_1[FSharpList[tuple[DataType, Any]]]) -> OptionalSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]:
    return OptionalSource_1__ctor_2B595(wx2)


def CellBuilder__Combine_168D7AF9(this: CellBuilder, wx1: SheetEntity_1[FSharpList[tuple[DataType, Any]]], wx2: OptionalSource_1[None]) -> OptionalSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]:
    return OptionalSource_1__ctor_2B595(wx1)


def CellBuilder__Combine_32507AAF(this: CellBuilder, wx1: RequiredSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]], wx2: SheetEntity_1[FSharpList[tuple[DataType, Any]]]) -> RequiredSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]:
    return RequiredSource_1__ctor_2B595(CellBuilder__Combine_F07E260(this, RequiredSource_1__get_Source(wx1), wx2))


def CellBuilder__Combine_Z3ABF6771(this: CellBuilder, wx1: SheetEntity_1[FSharpList[tuple[DataType, Any]]], wx2: RequiredSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]) -> RequiredSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]:
    return RequiredSource_1__ctor_2B595(CellBuilder__Combine_F07E260(this, wx1, RequiredSource_1__get_Source(wx2)))


def CellBuilder__Combine_Z2BFC83F8(this: CellBuilder, wx1: OptionalSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]], wx2: SheetEntity_1[FSharpList[tuple[DataType, Any]]]) -> OptionalSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]:
    return OptionalSource_1__ctor_2B595(CellBuilder__Combine_F07E260(this, OptionalSource_1__get_Source(wx1), wx2))


def CellBuilder__Combine_245D6C8(this: CellBuilder, wx1: SheetEntity_1[FSharpList[tuple[DataType, Any]]], wx2: OptionalSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]) -> OptionalSource_1[SheetEntity_1[FSharpList[tuple[DataType, Any]]]]:
    return OptionalSource_1__ctor_2B595(CellBuilder__Combine_F07E260(this, wx1, OptionalSource_1__get_Source(wx2)))


def CellBuilder__AsCellElement_6F87C2ED(this: CellBuilder, children: SheetEntity_1[FSharpList[tuple[DataType, Any]]]) -> SheetEntity_1[tuple[tuple[DataType, Any], int | None]]:
    (pattern_matching_result, messages, v, messages_1, vals, messages_2, messages_3) = (None, None, None, None, None, None, None)
    if children.tag == 2:
        pattern_matching_result = 2
        messages_2 = children.fields[0]

    elif children.tag == 1:
        pattern_matching_result = 3
        messages_3 = children.fields[0]

    elif not is_empty(children.fields[0]):
        if is_empty(tail(children.fields[0])):
            pattern_matching_result = 0
            messages = children.fields[1]
            v = head(children.fields[0])

        else: 
            pattern_matching_result = 1
            messages_1 = children.fields[1]
            vals = children.fields[0]


    else: 
        pattern_matching_result = 1
        messages_1 = children.fields[1]
        vals = children.fields[0]

    if pattern_matching_result == 0:
        return SheetEntity_1(0, (v, None), messages)

    elif pattern_matching_result == 1:
        return SheetEntity_1(0, (ReduceOperation__Reduce_70D1A69E(this.reducer, vals), None), messages_1)

    elif pattern_matching_result == 2:
        return SheetEntity_1(2, messages_2)

    elif pattern_matching_result == 3:
        return SheetEntity_1(1, messages_3)



__all__ = ["ReduceOperation_reflection", "ReduceOperation__Reduce_70D1A69E", "CellBuilder_reflection", "CellBuilder_get_Empty", "CellBuilder__SignMessages_3F89FFFC", "CellBuilder__Yield_36B9E420", "CellBuilder__Combine_F07E260", "CellBuilder__Combine_4BCE267E", "CellBuilder__Combine_Z2CF9C142", "CellBuilder__Combine_583F6A19", "CellBuilder__Combine_168D7AF9", "CellBuilder__Combine_32507AAF", "CellBuilder__Combine_Z3ABF6771", "CellBuilder__Combine_Z2BFC83F8", "CellBuilder__Combine_245D6C8", "CellBuilder__AsCellElement_6F87C2ED"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from ..fable_library.array_ import (append, map)
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, array_type, record_type)
from ..fable_library.types import (Array, Record)
from .commit import (Commit, Commit_reflection)
from .file_system_tree import (FileSystemTree, FileSystemTree_reflection)

def _expr2() -> TypeInfo:
    return record_type("ARCtrl.FileSystem.FileSystem", [], FileSystem, lambda: [("Tree", FileSystemTree_reflection()), ("History", array_type(Commit_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class FileSystem(Record):
    Tree: FileSystemTree
    History: Array[Commit]
    @staticmethod
    def create(tree: FileSystemTree, history: Array[Commit] | None=None) -> FileSystem:
        return FileSystem(tree, default_arg(history, []))

    def AddFile(self, path: str) -> FileSystem:
        this: FileSystem = self
        return FileSystem(this.Tree.AddFile(path), this.History)

    @staticmethod
    def add_file(path: str) -> Callable[[FileSystem], FileSystem]:
        def _arrow1(fs: FileSystem) -> FileSystem:
            return fs.AddFile(path)

        return _arrow1

    @staticmethod
    def from_file_paths(paths: Array[str]) -> FileSystem:
        tree: FileSystemTree = FileSystemTree.from_file_paths(paths)
        return FileSystem.create(tree = tree)

    def Union(self, other: FileSystem) -> FileSystem:
        this: FileSystem = self
        tree: FileSystemTree = this.Tree.Union(other.Tree)
        history: Array[Commit] = append(this.History, other.History, None)
        return FileSystem.create(tree = tree, history = history)

    def Copy(self, __unit: None=None) -> FileSystem:
        this: FileSystem = self
        fst_copy: FileSystemTree = this.Tree.Copy()
        def mapping(x: Commit) -> Commit:
            return x

        history_copy: Array[Commit] = map(mapping, this.History, None)
        return FileSystem.create(tree = fst_copy, history = history_copy)


FileSystem_reflection = _expr2

__all__ = ["FileSystem_reflection"]




from __future__ import annotations
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.string_ import (to_text, printf, to_console)
from ...fable_library.types import (Array, uint32)
from ..fs_address import (FsAddress, CellReference_toIndices, CellReference_moveHorizontal)

def _expr203() -> TypeInfo:
    return class_type("FsSpreadsheet.FsRangeAddress", None, FsRangeAddress)


class FsRangeAddress:
    def __init__(self, first_address: FsAddress, last_address: FsAddress) -> None:
        self.first_address: FsAddress = first_address
        self.last_address: FsAddress = last_address
        self._firstAddress: FsAddress = self.first_address
        self._lastAddress: FsAddress = self.last_address

    @staticmethod
    def from_string(range_address: str) -> FsRangeAddress:
        pattern_input: tuple[str, str] = Range_toBoundaries(range_address)
        return FsRangeAddress(FsAddress.from_string(pattern_input[0]), FsAddress.from_string(pattern_input[1]))

    def Copy(self, __unit: None=None) -> FsRangeAddress:
        self_1: FsRangeAddress = self
        return FsRangeAddress.from_string(self_1.Range)

    @staticmethod
    def copy(range_address: FsRangeAddress) -> FsRangeAddress:
        return range_address.Copy()

    def Extend(self, address: FsAddress) -> None:
        self_1: FsRangeAddress = self
        if address.RowNumber < self_1._firstAddress.RowNumber:
            self_1._firstAddress.RowNumber = address.RowNumber or 0

        if address.RowNumber > self_1._lastAddress.RowNumber:
            self_1._lastAddress.RowNumber = address.RowNumber or 0

        if address.ColumnNumber < self_1._firstAddress.ColumnNumber:
            self_1._firstAddress.ColumnNumber = address.ColumnNumber or 0

        if address.ColumnNumber > self_1._lastAddress.ColumnNumber:
            self_1._lastAddress.ColumnNumber = address.ColumnNumber or 0


    def Normalize(self, __unit: None=None) -> None:
        self_1: FsRangeAddress = self
        pattern_input: tuple[int, int] = ((self_1.first_address.RowNumber, self_1.last_address.RowNumber)) if (self_1.first_address.RowNumber < self_1.last_address.RowNumber) else ((self_1.last_address.RowNumber, self_1.first_address.RowNumber))
        pattern_input_1: tuple[int, int] = ((self_1.first_address.RowNumber, self_1.last_address.RowNumber)) if (self_1.first_address.RowNumber < self_1.last_address.RowNumber) else ((self_1.last_address.RowNumber, self_1.first_address.RowNumber))
        self_1._firstAddress = FsAddress(pattern_input[0], pattern_input_1[0])
        self_1._lastAddress = FsAddress(pattern_input[1], pattern_input_1[1])

    @property
    def Range(self, __unit: None=None) -> str:
        self_1: FsRangeAddress = self
        return Range_ofBoundaries(self_1._firstAddress.Address, self_1._lastAddress.Address)

    @Range.setter
    def Range(self, address: str) -> None:
        self_1: FsRangeAddress = self
        pattern_input: tuple[str, str] = Range_toBoundaries(address)
        self_1._firstAddress = FsAddress.from_string(pattern_input[0])
        self_1._lastAddress = FsAddress.from_string(pattern_input[1])

    def __str__(self, __unit: None=None) -> str:
        self_1: FsRangeAddress = self
        return self_1.Range

    @property
    def FirstAddress(self, __unit: None=None) -> FsAddress:
        self_1: FsRangeAddress = self
        return self_1._firstAddress

    @property
    def LastAddress(self, __unit: None=None) -> FsAddress:
        self_1: FsRangeAddress = self
        return self_1._lastAddress

    def Union(self, range_address: FsRangeAddress) -> FsRangeAddress:
        self_1: FsRangeAddress = self
        self_1.Extend(range_address.FirstAddress)
        self_1.Extend(range_address.LastAddress)
        return self_1


FsRangeAddress_reflection = _expr203

def FsRangeAddress__ctor_7E77A4A0(first_address: FsAddress, last_address: FsAddress) -> FsRangeAddress:
    return FsRangeAddress(first_address, last_address)


def Range_ofBoundaries(from_cell_reference: str, to_cell_reference: str) -> str:
    return to_text(printf("%s:%s"))(from_cell_reference)(to_cell_reference)


def Range_toBoundaries(area: str) -> tuple[str, str]:
    a: Array[str] = area.split(":")
    return (a[0], a[1])


def Range_rightBoundary(area: str) -> uint32:
    return CellReference_toIndices(Range_toBoundaries(area)[1])[0]


def Range_leftBoundary(area: str) -> uint32:
    return CellReference_toIndices(Range_toBoundaries(area)[0])[0]


def Range_upperBoundary(area: str) -> uint32:
    return CellReference_toIndices(Range_toBoundaries(area)[0])[1]


def Range_lowerBoundary(area: str) -> uint32:
    return CellReference_toIndices(Range_toBoundaries(area)[1])[1]


def Range_moveHorizontal(amount: int, area: str) -> str:
    tupled_arg_1: tuple[str, str]
    tupled_arg: tuple[str, str] = Range_toBoundaries(area)
    tupled_arg_1 = (CellReference_moveHorizontal(amount, tupled_arg[0]), CellReference_moveHorizontal(amount, tupled_arg[1]))
    return Range_ofBoundaries(tupled_arg_1[0], tupled_arg_1[1])


def Range_moveVertical(amount: int, area: str) -> str:
    tupled_arg_1: tuple[str, str]
    tupled_arg: tuple[str, str] = Range_toBoundaries(area)
    tupled_arg_1 = (CellReference_moveHorizontal(amount, tupled_arg[0]), CellReference_moveHorizontal(amount, tupled_arg[1]))
    return Range_ofBoundaries(tupled_arg_1[0], tupled_arg_1[1])


def Range_extendRight(amount: int, area: str) -> str:
    tupled_arg_1: tuple[str, str]
    tupled_arg: tuple[str, str] = Range_toBoundaries(area)
    tupled_arg_1 = (tupled_arg[0], CellReference_moveHorizontal(amount, tupled_arg[1]))
    return Range_ofBoundaries(tupled_arg_1[0], tupled_arg_1[1])


def Range_extendLeft(amount: int, area: str) -> str:
    tupled_arg_1: tuple[str, str]
    tupled_arg: tuple[str, str] = Range_toBoundaries(area)
    tupled_arg_1 = (CellReference_moveHorizontal(amount, tupled_arg[0]), tupled_arg[1])
    return Range_ofBoundaries(tupled_arg_1[0], tupled_arg_1[1])


def Range_referenceExceedsAreaRight(reference: str, area: str) -> bool:
    return CellReference_toIndices(reference)[0] > Range_rightBoundary(area)


def Range_referenceExceedsAreaLeft(reference: str, area: str) -> bool:
    return CellReference_toIndices(reference)[0] < Range_leftBoundary(area)


def Range_referenceExceedsAreaAbove(reference: str, area: str) -> bool:
    return CellReference_toIndices(reference)[1] > Range_upperBoundary(area)


def Range_referenceExceedsAreaBelow(reference: str, area: str) -> bool:
    return CellReference_toIndices(reference)[1] < Range_lowerBoundary(area)


def Range_referenceExceedsArea(reference: str, area: str) -> bool:
    if True if (True if Range_referenceExceedsAreaRight(reference, area) else Range_referenceExceedsAreaLeft(reference, area)) else Range_referenceExceedsAreaAbove(reference, area):
        return True

    else: 
        return Range_referenceExceedsAreaBelow(reference, area)



def Range_isCorrect(area: str) -> bool:
    try: 
        hor: bool = Range_leftBoundary(area) <= Range_rightBoundary(area)
        ver: bool = Range_upperBoundary(area) <= Range_lowerBoundary(area)
        if not hor:
            to_console(printf("Right area boundary must be higher or equal to left area boundary."))

        if not ver:
            to_console(printf("Lower area boundary must be higher or equal to upper area boundary."))

        return ver if hor else False

    except Exception as err:
        arg_1: str = str(err)
        to_console(printf("Area \"%s\" could not be parsed: %s"))(area)(arg_1)
        return False



__all__ = ["FsRangeAddress_reflection", "Range_ofBoundaries", "Range_toBoundaries", "Range_rightBoundary", "Range_leftBoundary", "Range_upperBoundary", "Range_lowerBoundary", "Range_moveHorizontal", "Range_moveVertical", "Range_extendRight", "Range_extendLeft", "Range_referenceExceedsAreaRight", "Range_referenceExceedsAreaLeft", "Range_referenceExceedsAreaAbove", "Range_referenceExceedsAreaBelow", "Range_referenceExceedsArea", "Range_isCorrect"]




from __future__ import annotations
from abc import abstractmethod
from collections.abc import Callable
from typing import (Any, Protocol, TypeVar)
from ..fable_library.boolean import try_parse as try_parse_2
from ..fable_library.date import (min_value, try_parse as try_parse_3)
from ..fable_library.date_offset import (min_value as min_value_1, try_parse as try_parse_4)
from ..fable_library.double import try_parse as try_parse_1
from ..fable_library.int32 import try_parse
from ..fable_library.list import (is_empty, head, tail, map as map_1, FSharpList, to_array, try_find, filter, exists, iterate)
from ..fable_library.map import (of_list, of_seq)
from ..fable_library.map_util import add_to_dict
from ..fable_library.mutable_map import Dictionary
from ..fable_library.option import (some, map as map_3, value as value_1)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import map as map_2
from ..fable_library.string_ import (to_text, printf)
from ..fable_library.types import (FSharpRef, Array)
from ..fable_library.util import (compare, equals as equals_1, structural_hash, IEnumerable_1)
from .Interop.py_interop import is_string
from .reader import read as read_1
from .yamlicious_types import YAMLElement

_A = TypeVar("_A")

_B = TypeVar("_B")

_A_ = TypeVar("_A_")

_C = TypeVar("_C")

_D = TypeVar("_D")

_E = TypeVar("_E")

_F = TypeVar("_F")

_G = TypeVar("_G")

_H = TypeVar("_H")

__A_ = TypeVar("__A_")

def Helper_isString(value: str) -> bool:
    return is_string(value)


def int_1(value: YAMLElement) -> int:
    (pattern_matching_result, v) = (None, None)
    if value.tag == 1:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 1:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        match_value: tuple[bool, int]
        out_arg: int = 0
        def _arrow490(__unit: None=None, value: Any=value) -> int:
            return out_arg

        def _arrow491(v_1: int, value: Any=value) -> None:
            nonlocal out_arg
            out_arg = v_1 or 0

        match_value = (try_parse(v.Value, 511, False, 32, FSharpRef(_arrow490, _arrow491)), out_arg)
        if match_value[0]:
            return match_value[1]

        else: 
            raise Exception((to_text(printf("%s: %A"))("Expected an int")(v.Value) + "\\nParameter name: ") + "value")


    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected an YAMLElement.Value")(value) + "\\nParameter name: ") + "value")



def float_1(value: YAMLElement) -> float:
    Culture: Any = {}
    (pattern_matching_result, v) = (None, None)
    if value.tag == 1:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 1:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        match_value: tuple[bool, float]
        out_arg: float = 0.0
        def _arrow492(__unit: None=None, value: Any=value) -> float:
            return out_arg

        def _arrow493(v_1: float, value: Any=value) -> None:
            nonlocal out_arg
            out_arg = v_1

        match_value = (try_parse_1(v.Value, FSharpRef(_arrow492, _arrow493)), out_arg)
        if match_value[0]:
            return match_value[1]

        else: 
            raise Exception((to_text(printf("%s: %A"))("Expected an int")(v.Value) + "\\nParameter name: ") + "value")


    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected an YAMLElement.Value")(value) + "\\nParameter name: ") + "value")



def char(value: YAMLElement) -> str:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 1:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 1:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        if Helper_isString(v.Value):
            return v.Value[0]

        else: 
            raise Exception((to_text(printf("%s: %A"))("Expected a char")(v.Value) + "\\nParameter name: ") + "value")


    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a char")(any_else) + "\\nParameter name: ") + "value")



def bool_1(value: YAMLElement) -> bool:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 1:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 1:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        match_value: tuple[bool, bool]
        out_arg: bool = False
        def _arrow494(__unit: None=None, value: Any=value) -> bool:
            return out_arg

        def _arrow495(v_1: bool, value: Any=value) -> None:
            nonlocal out_arg
            out_arg = v_1

        match_value = (try_parse_2(v.Value, FSharpRef(_arrow494, _arrow495)), out_arg)
        if match_value[0]:
            return match_value[1]

        else: 
            raise Exception((to_text(printf("%s: %A"))("Expected a bool")(v.Value) + "\\nParameter name: ") + "value")


    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a bool")(any_else) + "\\nParameter name: ") + "value")



def map(key_decoder: Callable[[str], _A], value_decoder: Callable[[YAMLElement], _B], value: YAMLElement) -> Any:
    if value.tag == 3:
        def mapping(x: YAMLElement, key_decoder: Any=key_decoder, value_decoder: Any=value_decoder, value: Any=value) -> tuple[_A, _B]:
            if x.tag == 0:
                return (key_decoder(x.fields[0].Value), value_decoder(x.fields[1]))

            else: 
                raise Exception((to_text(printf("%s: %A"))("Expected a mapping")(x) + "\\nParameter name: ") + "value")


        class ObjectExpr496:
            @property
            def Compare(self) -> Callable[[_A_, _A_], int]:
                return compare

        return of_list(map_1(mapping, value.fields[0]), ObjectExpr496())

    else: 
        raise Exception((to_text(printf("%s: %A"))("Expected a map")(value) + "\\nParameter name: ") + "value")



def dict_1(key_decoder: Callable[[str], _A], value_decoder: Callable[[YAMLElement], _B], value: YAMLElement) -> Any:
    if value.tag == 3:
        def mapping(x: YAMLElement, key_decoder: Any=key_decoder, value_decoder: Any=value_decoder, value: Any=value) -> tuple[_A, _B]:
            if x.tag == 0:
                return (key_decoder(x.fields[0].Value), value_decoder(x.fields[1]))

            else: 
                raise Exception((to_text(printf("%s: %A"))("Expected a mapping")(x) + "\\nParameter name: ") + "value")


        class ObjectExpr497:
            @property
            def Compare(self) -> Callable[[_A_, _A_], int]:
                return compare

        class ObjectExpr498:
            @property
            def Equals(self) -> Callable[[_A_, _A_], bool]:
                return equals_1

            @property
            def GetHashCode(self) -> Callable[[_A_], int]:
                return structural_hash

        return Dictionary(of_seq(map_1(mapping, value.fields[0]), ObjectExpr497()), ObjectExpr498())

    else: 
        raise Exception((to_text(printf("%s: %A"))("Expected a dictionary")(value) + "\\nParameter name: ") + "value")



def datetime(value: YAMLElement) -> Any:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 1:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 1:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        match_value: tuple[bool, Any]
        out_arg: Any = min_value()
        def _arrow499(__unit: None=None, value: Any=value) -> Any:
            return out_arg

        def _arrow500(v_1: Any, value: Any=value) -> None:
            nonlocal out_arg
            out_arg = v_1

        match_value = (try_parse_3(v.Value, FSharpRef(_arrow499, _arrow500)), out_arg)
        if match_value[0]:
            return match_value[1]

        else: 
            raise Exception((to_text(printf("%s: %A"))("Expected a DateTime")(v.Value) + "\\nParameter name: ") + "value")


    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a DateTime")(any_else) + "\\nParameter name: ") + "value")



def datetime_offset(value: YAMLElement) -> Any:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 1:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 1:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        match_value: tuple[bool, Any]
        out_arg: Any = min_value_1()
        def _arrow501(__unit: None=None, value: Any=value) -> Any:
            return out_arg

        def _arrow502(v_1: Any, value: Any=value) -> None:
            nonlocal out_arg
            out_arg = v_1

        match_value = (try_parse_4(v.Value, FSharpRef(_arrow501, _arrow502)), out_arg)
        if match_value[0]:
            return match_value[1]

        else: 
            raise Exception((to_text(printf("%s: %A"))("Expected a DateTimeOffset")(v.Value) + "\\nParameter name: ") + "value")


    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a DateTimeOffset")(any_else) + "\\nParameter name: ") + "value")



def option(decoder: Callable[[YAMLElement], _A], value: YAMLElement) -> Any | None:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 1:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 1:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        if v.Value == "null":
            return None

        else: 
            return some(decoder(value))


    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected an option")(any_else) + "\\nParameter name: ") + "value")



def string(value: YAMLElement) -> str:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 1:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 1:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return v.Value

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a string")(any_else) + "\\nParameter name: ") + "value")



def tuple2(decoder_a: Callable[[YAMLElement], _A], decoder_b: Callable[[YAMLElement], _B], value: YAMLElement) -> tuple[_A, _B]:
    (pattern_matching_result, a, b, any_else) = (None, None, None, None)
    if value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if not is_empty(head(value.fields[0]).fields[0]):
                    if not is_empty(tail(head(value.fields[0]).fields[0])):
                        if is_empty(tail(tail(head(value.fields[0]).fields[0]))):
                            if is_empty(tail(value.fields[0])):
                                pattern_matching_result = 0
                                a = head(head(value.fields[0]).fields[0])
                                b = head(tail(head(value.fields[0]).fields[0]))

                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    elif value.tag == 2:
        if not is_empty(value.fields[0]):
            if not is_empty(tail(value.fields[0])):
                if is_empty(tail(tail(value.fields[0]))):
                    pattern_matching_result = 0
                    a = head(value.fields[0])
                    b = head(tail(value.fields[0]))

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return (decoder_a(a), decoder_b(b))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a tuple2")(any_else) + "\\nParameter name: ") + "value")



def tuple3(decoder_a: Callable[[YAMLElement], _A], decoder_b: Callable[[YAMLElement], _B], decoder_c: Callable[[YAMLElement], _C], value: YAMLElement) -> tuple[_A, _B, _C]:
    (pattern_matching_result, a, b, c, any_else) = (None, None, None, None, None)
    if value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if not is_empty(head(value.fields[0]).fields[0]):
                    if not is_empty(tail(head(value.fields[0]).fields[0])):
                        if not is_empty(tail(tail(head(value.fields[0]).fields[0]))):
                            if is_empty(tail(tail(tail(head(value.fields[0]).fields[0])))):
                                if is_empty(tail(value.fields[0])):
                                    pattern_matching_result = 0
                                    a = head(head(value.fields[0]).fields[0])
                                    b = head(tail(head(value.fields[0]).fields[0]))
                                    c = head(tail(tail(head(value.fields[0]).fields[0])))

                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    elif value.tag == 2:
        if not is_empty(value.fields[0]):
            if not is_empty(tail(value.fields[0])):
                if not is_empty(tail(tail(value.fields[0]))):
                    if is_empty(tail(tail(tail(value.fields[0])))):
                        pattern_matching_result = 0
                        a = head(value.fields[0])
                        b = head(tail(value.fields[0]))
                        c = head(tail(tail(value.fields[0])))

                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return (decoder_a(a), decoder_b(b), decoder_c(c))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a tuple3")(any_else) + "\\nParameter name: ") + "value")



def tuple4(decoder_a: Callable[[YAMLElement], _A], decoder_b: Callable[[YAMLElement], _B], decoder_c: Callable[[YAMLElement], _C], decoder_d: Callable[[YAMLElement], _D], value: YAMLElement) -> tuple[_A, _B, _C, _D]:
    (pattern_matching_result, a, b, c, d, any_else) = (None, None, None, None, None, None)
    if value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if not is_empty(head(value.fields[0]).fields[0]):
                    if not is_empty(tail(head(value.fields[0]).fields[0])):
                        if not is_empty(tail(tail(head(value.fields[0]).fields[0]))):
                            if not is_empty(tail(tail(tail(head(value.fields[0]).fields[0])))):
                                if is_empty(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))):
                                    if is_empty(tail(value.fields[0])):
                                        pattern_matching_result = 0
                                        a = head(head(value.fields[0]).fields[0])
                                        b = head(tail(head(value.fields[0]).fields[0]))
                                        c = head(tail(tail(head(value.fields[0]).fields[0])))
                                        d = head(tail(tail(tail(head(value.fields[0]).fields[0]))))

                                    else: 
                                        pattern_matching_result = 1
                                        any_else = value


                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    elif value.tag == 2:
        if not is_empty(value.fields[0]):
            if not is_empty(tail(value.fields[0])):
                if not is_empty(tail(tail(value.fields[0]))):
                    if not is_empty(tail(tail(tail(value.fields[0])))):
                        if is_empty(tail(tail(tail(tail(value.fields[0]))))):
                            pattern_matching_result = 0
                            a = head(value.fields[0])
                            b = head(tail(value.fields[0]))
                            c = head(tail(tail(value.fields[0])))
                            d = head(tail(tail(tail(value.fields[0]))))

                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return (decoder_a(a), decoder_b(b), decoder_c(c), decoder_d(d))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a tuple4")(any_else) + "\\nParameter name: ") + "value")



def tuple5(decoder_a: Callable[[YAMLElement], _A], decoder_b: Callable[[YAMLElement], _B], decoder_c: Callable[[YAMLElement], _C], decoder_d: Callable[[YAMLElement], _D], decoder_e: Callable[[YAMLElement], _E], value: YAMLElement) -> tuple[_A, _B, _C, _D, _E]:
    (pattern_matching_result, a, b, c, d, e, any_else) = (None, None, None, None, None, None, None)
    if value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if not is_empty(head(value.fields[0]).fields[0]):
                    if not is_empty(tail(head(value.fields[0]).fields[0])):
                        if not is_empty(tail(tail(head(value.fields[0]).fields[0]))):
                            if not is_empty(tail(tail(tail(head(value.fields[0]).fields[0])))):
                                if not is_empty(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))):
                                    if is_empty(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))):
                                        if is_empty(tail(value.fields[0])):
                                            pattern_matching_result = 0
                                            a = head(head(value.fields[0]).fields[0])
                                            b = head(tail(head(value.fields[0]).fields[0]))
                                            c = head(tail(tail(head(value.fields[0]).fields[0])))
                                            d = head(tail(tail(tail(head(value.fields[0]).fields[0]))))
                                            e = head(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))

                                        else: 
                                            pattern_matching_result = 1
                                            any_else = value


                                    else: 
                                        pattern_matching_result = 1
                                        any_else = value


                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    elif value.tag == 2:
        if not is_empty(value.fields[0]):
            if not is_empty(tail(value.fields[0])):
                if not is_empty(tail(tail(value.fields[0]))):
                    if not is_empty(tail(tail(tail(value.fields[0])))):
                        if not is_empty(tail(tail(tail(tail(value.fields[0]))))):
                            if is_empty(tail(tail(tail(tail(tail(value.fields[0])))))):
                                pattern_matching_result = 0
                                a = head(value.fields[0])
                                b = head(tail(value.fields[0]))
                                c = head(tail(tail(value.fields[0])))
                                d = head(tail(tail(tail(value.fields[0]))))
                                e = head(tail(tail(tail(tail(value.fields[0])))))

                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return (decoder_a(a), decoder_b(b), decoder_c(c), decoder_d(d), decoder_e(e))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a tuple5")(any_else) + "\\nParameter name: ") + "value")



def tuple6(decoder_a: Callable[[YAMLElement], _A], decoder_b: Callable[[YAMLElement], _B], decoder_c: Callable[[YAMLElement], _C], decoder_d: Callable[[YAMLElement], _D], decoder_e: Callable[[YAMLElement], _E], decoder_f: Callable[[YAMLElement], _F], value: YAMLElement) -> tuple[_A, _B, _C, _D, _E, _F]:
    (pattern_matching_result, a, b, c, d, e, f, any_else) = (None, None, None, None, None, None, None, None)
    if value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if not is_empty(head(value.fields[0]).fields[0]):
                    if not is_empty(tail(head(value.fields[0]).fields[0])):
                        if not is_empty(tail(tail(head(value.fields[0]).fields[0]))):
                            if not is_empty(tail(tail(tail(head(value.fields[0]).fields[0])))):
                                if not is_empty(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))):
                                    if not is_empty(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))):
                                        if is_empty(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))))):
                                            if is_empty(tail(value.fields[0])):
                                                pattern_matching_result = 0
                                                a = head(head(value.fields[0]).fields[0])
                                                b = head(tail(head(value.fields[0]).fields[0]))
                                                c = head(tail(tail(head(value.fields[0]).fields[0])))
                                                d = head(tail(tail(tail(head(value.fields[0]).fields[0]))))
                                                e = head(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))
                                                f = head(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))))

                                            else: 
                                                pattern_matching_result = 1
                                                any_else = value


                                        else: 
                                            pattern_matching_result = 1
                                            any_else = value


                                    else: 
                                        pattern_matching_result = 1
                                        any_else = value


                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    elif value.tag == 2:
        if not is_empty(value.fields[0]):
            if not is_empty(tail(value.fields[0])):
                if not is_empty(tail(tail(value.fields[0]))):
                    if not is_empty(tail(tail(tail(value.fields[0])))):
                        if not is_empty(tail(tail(tail(tail(value.fields[0]))))):
                            if not is_empty(tail(tail(tail(tail(tail(value.fields[0])))))):
                                if is_empty(tail(tail(tail(tail(tail(tail(value.fields[0]))))))):
                                    pattern_matching_result = 0
                                    a = head(value.fields[0])
                                    b = head(tail(value.fields[0]))
                                    c = head(tail(tail(value.fields[0])))
                                    d = head(tail(tail(tail(value.fields[0]))))
                                    e = head(tail(tail(tail(tail(value.fields[0])))))
                                    f = head(tail(tail(tail(tail(tail(value.fields[0]))))))

                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return (decoder_a(a), decoder_b(b), decoder_c(c), decoder_d(d), decoder_e(e), decoder_f(f))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a tuple6")(any_else) + "\\nParameter name: ") + "value")



def tuple7(decoder_a: Callable[[YAMLElement], _A], decoder_b: Callable[[YAMLElement], _B], decoder_c: Callable[[YAMLElement], _C], decoder_d: Callable[[YAMLElement], _D], decoder_e: Callable[[YAMLElement], _E], decoder_f: Callable[[YAMLElement], _F], decoder_g: Callable[[YAMLElement], _G], value: YAMLElement) -> tuple[_A, _B, _C, _D, _E, _F, _G]:
    (pattern_matching_result, a, b, c, d, e, f, g, any_else) = (None, None, None, None, None, None, None, None, None)
    if value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if not is_empty(head(value.fields[0]).fields[0]):
                    if not is_empty(tail(head(value.fields[0]).fields[0])):
                        if not is_empty(tail(tail(head(value.fields[0]).fields[0]))):
                            if not is_empty(tail(tail(tail(head(value.fields[0]).fields[0])))):
                                if not is_empty(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))):
                                    if not is_empty(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))):
                                        if not is_empty(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))))):
                                            if is_empty(tail(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))))):
                                                if is_empty(tail(value.fields[0])):
                                                    pattern_matching_result = 0
                                                    a = head(head(value.fields[0]).fields[0])
                                                    b = head(tail(head(value.fields[0]).fields[0]))
                                                    c = head(tail(tail(head(value.fields[0]).fields[0])))
                                                    d = head(tail(tail(tail(head(value.fields[0]).fields[0]))))
                                                    e = head(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))
                                                    f = head(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))))
                                                    g = head(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))))

                                                else: 
                                                    pattern_matching_result = 1
                                                    any_else = value


                                            else: 
                                                pattern_matching_result = 1
                                                any_else = value


                                        else: 
                                            pattern_matching_result = 1
                                            any_else = value


                                    else: 
                                        pattern_matching_result = 1
                                        any_else = value


                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    elif value.tag == 2:
        if not is_empty(value.fields[0]):
            if not is_empty(tail(value.fields[0])):
                if not is_empty(tail(tail(value.fields[0]))):
                    if not is_empty(tail(tail(tail(value.fields[0])))):
                        if not is_empty(tail(tail(tail(tail(value.fields[0]))))):
                            if not is_empty(tail(tail(tail(tail(tail(value.fields[0])))))):
                                if not is_empty(tail(tail(tail(tail(tail(tail(value.fields[0]))))))):
                                    if is_empty(tail(tail(tail(tail(tail(tail(tail(value.fields[0])))))))):
                                        pattern_matching_result = 0
                                        a = head(value.fields[0])
                                        b = head(tail(value.fields[0]))
                                        c = head(tail(tail(value.fields[0])))
                                        d = head(tail(tail(tail(value.fields[0]))))
                                        e = head(tail(tail(tail(tail(value.fields[0])))))
                                        f = head(tail(tail(tail(tail(tail(value.fields[0]))))))
                                        g = head(tail(tail(tail(tail(tail(tail(value.fields[0])))))))

                                    else: 
                                        pattern_matching_result = 1
                                        any_else = value


                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return (decoder_a(a), decoder_b(b), decoder_c(c), decoder_d(d), decoder_e(e), decoder_f(f), decoder_g(g))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a tuple7")(any_else) + "\\nParameter name: ") + "value")



def tuple8(decoder_a: Callable[[YAMLElement], _A], decoder_b: Callable[[YAMLElement], _B], decoder_c: Callable[[YAMLElement], _C], decoder_d: Callable[[YAMLElement], _D], decoder_e: Callable[[YAMLElement], _E], decoder_f: Callable[[YAMLElement], _F], decoder_g: Callable[[YAMLElement], _G], decoder_h: Callable[[YAMLElement], _H], value: YAMLElement) -> tuple[_A, _B, _C, _D, _E, _F, _G, _H]:
    (pattern_matching_result, a, b, c, d, e, f, g, h, any_else) = (None, None, None, None, None, None, None, None, None, None)
    if value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if not is_empty(head(value.fields[0]).fields[0]):
                    if not is_empty(tail(head(value.fields[0]).fields[0])):
                        if not is_empty(tail(tail(head(value.fields[0]).fields[0]))):
                            if not is_empty(tail(tail(tail(head(value.fields[0]).fields[0])))):
                                if not is_empty(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))):
                                    if not is_empty(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))):
                                        if not is_empty(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))))):
                                            if not is_empty(tail(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))))):
                                                if is_empty(tail(tail(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))))))):
                                                    if is_empty(tail(value.fields[0])):
                                                        pattern_matching_result = 0
                                                        a = head(head(value.fields[0]).fields[0])
                                                        b = head(tail(head(value.fields[0]).fields[0]))
                                                        c = head(tail(tail(head(value.fields[0]).fields[0])))
                                                        d = head(tail(tail(tail(head(value.fields[0]).fields[0]))))
                                                        e = head(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))
                                                        f = head(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))))
                                                        g = head(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0])))))))
                                                        h = head(tail(tail(tail(tail(tail(tail(tail(head(value.fields[0]).fields[0]))))))))

                                                    else: 
                                                        pattern_matching_result = 1
                                                        any_else = value


                                                else: 
                                                    pattern_matching_result = 1
                                                    any_else = value


                                            else: 
                                                pattern_matching_result = 1
                                                any_else = value


                                        else: 
                                            pattern_matching_result = 1
                                            any_else = value


                                    else: 
                                        pattern_matching_result = 1
                                        any_else = value


                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    elif value.tag == 2:
        if not is_empty(value.fields[0]):
            if not is_empty(tail(value.fields[0])):
                if not is_empty(tail(tail(value.fields[0]))):
                    if not is_empty(tail(tail(tail(value.fields[0])))):
                        if not is_empty(tail(tail(tail(tail(value.fields[0]))))):
                            if not is_empty(tail(tail(tail(tail(tail(value.fields[0])))))):
                                if not is_empty(tail(tail(tail(tail(tail(tail(value.fields[0]))))))):
                                    if not is_empty(tail(tail(tail(tail(tail(tail(tail(value.fields[0])))))))):
                                        if is_empty(tail(tail(tail(tail(tail(tail(tail(tail(value.fields[0]))))))))):
                                            pattern_matching_result = 0
                                            a = head(value.fields[0])
                                            b = head(tail(value.fields[0]))
                                            c = head(tail(tail(value.fields[0])))
                                            d = head(tail(tail(tail(value.fields[0]))))
                                            e = head(tail(tail(tail(tail(value.fields[0])))))
                                            f = head(tail(tail(tail(tail(tail(value.fields[0]))))))
                                            g = head(tail(tail(tail(tail(tail(tail(value.fields[0])))))))
                                            h = head(tail(tail(tail(tail(tail(tail(tail(value.fields[0]))))))))

                                        else: 
                                            pattern_matching_result = 1
                                            any_else = value


                                    else: 
                                        pattern_matching_result = 1
                                        any_else = value


                                else: 
                                    pattern_matching_result = 1
                                    any_else = value


                            else: 
                                pattern_matching_result = 1
                                any_else = value


                        else: 
                            pattern_matching_result = 1
                            any_else = value


                    else: 
                        pattern_matching_result = 1
                        any_else = value


                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return (decoder_a(a), decoder_b(b), decoder_c(c), decoder_d(d), decoder_e(e), decoder_f(f), decoder_g(g), decoder_h(h))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a tuple8")(any_else) + "\\nParameter name: ") + "value")



def list_1(decoder: Callable[[YAMLElement], _A], value: YAMLElement) -> FSharpList[Any]:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 2:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return map_1(decoder, v)

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a list")(any_else) + "\\nParameter name: ") + "value")



def array(decoder: Callable[[YAMLElement], _A], value: YAMLElement) -> Array[Any]:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 2:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return to_array(map_1(decoder, v))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected an array")(any_else) + "\\nParameter name: ") + "value")



def seq(decoder: Callable[[YAMLElement], _A], value: YAMLElement) -> IEnumerable_1[Any]:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 2:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return map_2(decoder, v)

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a seq")(any_else) + "\\nParameter name: ") + "value")



def resizearray(decoder: Callable[[YAMLElement], _A], value: YAMLElement) -> Array[Any]:
    (pattern_matching_result, v, any_else) = (None, None, None)
    if value.tag == 2:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 1
                    any_else = value


            else: 
                pattern_matching_result = 1
                any_else = value


        else: 
            pattern_matching_result = 1
            any_else = value


    else: 
        pattern_matching_result = 1
        any_else = value

    if pattern_matching_result == 0:
        return list(map_1(decoder, v))

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a resizearray")(any_else) + "\\nParameter name: ") + "value")



def values(decoder: Callable[[YAMLElement], _A], value: YAMLElement) -> FSharpList[Any]:
    (pattern_matching_result, v, any_else_1) = (None, None, None)
    if value.tag == 2:
        pattern_matching_result = 0
        v = value.fields[0]

    elif value.tag == 3:
        if not is_empty(value.fields[0]):
            if head(value.fields[0]).tag == 2:
                if is_empty(tail(value.fields[0])):
                    pattern_matching_result = 0
                    v = head(value.fields[0]).fields[0]

                else: 
                    pattern_matching_result = 0
                    v = value.fields[0]


            else: 
                pattern_matching_result = 0
                v = value.fields[0]


        else: 
            pattern_matching_result = 0
            v = value.fields[0]


    else: 
        pattern_matching_result = 1
        any_else_1 = value

    if pattern_matching_result == 0:
        def mapping(_arg: YAMLElement, decoder: Any=decoder, value: Any=value) -> _A:
            if _arg.tag == 1:
                return decoder(_arg)

            else: 
                raise Exception((to_text(printf("%s: %A"))("Expected a values")(_arg) + "\\nParameter name: ") + "value")


        return map_1(mapping, v)

    elif pattern_matching_result == 1:
        raise Exception((to_text(printf("%s: %A"))("Expected a values")(any_else_1) + "\\nParameter name: ") + "value")



class IRequiredGetter(Protocol):
    @abstractmethod
    def Field(self, __arg0: str, __arg1: Callable[[YAMLElement], _A]) -> _A:
        ...


class IOptionalGetter(Protocol):
    @abstractmethod
    def Field(self, __arg0: str, __arg1: Callable[[YAMLElement], _A]) -> _A | None:
        ...


class IMultipleOptionalGetter(Protocol):
    @abstractmethod
    def FieldList(self, __arg0: FSharpList[str]) -> Any:
        ...


class IOverflowGetter(Protocol):
    @abstractmethod
    def FieldList(self, __arg0: FSharpList[str]) -> Any:
        ...


class IGetters(Protocol):
    @property
    @abstractmethod
    def MultipleOptional(self) -> IMultipleOptionalGetter:
        ...

    @property
    @abstractmethod
    def Optional(self) -> IOptionalGetter:
        ...

    @property
    @abstractmethod
    def Overflow(self) -> IOverflowGetter:
        ...

    @property
    @abstractmethod
    def Required(self) -> IRequiredGetter:
        ...


def _expr515() -> TypeInfo:
    return class_type("YAMLicious.Decode.Getter", None, Getter)


class Getter:
    def __init__(self, ele: YAMLElement) -> None:
        def _arrow508(__unit: None=None) -> IRequiredGetter:
            _this: Any = self
            class ObjectExpr507(IRequiredGetter):
                def Field(self, field_name: str, dec: Callable[[YAMLElement], __A_]) -> Any:
                    if ele.tag == 3:
                        def mapping(_arg_1: YAMLElement) -> __A_:
                            if _arg_1.tag == 0:
                                return dec(_arg_1.fields[1])

                            else: 
                                raise Exception((to_text(printf("%s: %A"))("Expected a mapping")(ele) + "\\nParameter name: ") + "value")


                        def predicate(_arg: YAMLElement) -> bool:
                            if _arg.tag == 0:
                                return _arg.fields[0].Value == field_name

                            else: 
                                return False


                        x: __A_ | None = map_3(mapping, try_find(predicate, ele.fields[0]))
                        if x is None:
                            def _arrow506(__unit: None=None) -> str:
                                arg_3: str = to_text(printf("Field not found: %s"))(field_name)
                                return to_text(printf("%s: %A"))(arg_3)(ele)

                            raise Exception((_arrow506() + "\\nParameter name: ") + "value")

                        else: 
                            return value_1(x)


                    else: 
                        raise Exception((to_text(printf("%s: %A"))("Expected an object")(ele) + "\\nParameter name: ") + "value")


            return ObjectExpr507()

        self.required: IRequiredGetter = _arrow508()
        def _arrow510(__unit: None=None) -> IOptionalGetter:
            _this_1: Any = self
            class ObjectExpr509(IOptionalGetter):
                def Field(self, field_name_1: str, dec_1: Callable[[YAMLElement], __A_]) -> Any | None:
                    if ele.tag == 3:
                        def mapping_1(_arg_3: YAMLElement) -> __A_:
                            if _arg_3.tag == 0:
                                return dec_1(_arg_3.fields[1])

                            else: 
                                raise Exception((to_text(printf("%s: %A"))("Expected a mapping")(ele) + "\\nParameter name: ") + "value")


                        def predicate_1(_arg_2: YAMLElement) -> bool:
                            if _arg_2.tag == 0:
                                return _arg_2.fields[0].Value == field_name_1

                            else: 
                                return False


                        return map_3(mapping_1, try_find(predicate_1, ele.fields[0]))

                    else: 
                        raise Exception((to_text(printf("%s: %A"))("Expected an object")(ele) + "\\nParameter name: ") + "value")


            return ObjectExpr509()

        self.optional: IOptionalGetter = _arrow510()
        def _arrow512(__unit: None=None) -> IMultipleOptionalGetter:
            _this_2: Any = self
            class ObjectExpr511(IMultipleOptionalGetter):
                def FieldList(self, field_names: FSharpList[str]) -> Any:
                    if ele.tag == 3:
                        def mapping_2(_arg_5: YAMLElement) -> tuple[str, YAMLElement]:
                            if _arg_5.tag == 0:
                                return (_arg_5.fields[0].Value, _arg_5.fields[1])

                            else: 
                                raise Exception((to_text(printf("%s: %A"))("Expected a mapping")(ele) + "\\nParameter name: ") + "value")


                        def predicate_3(_arg_4: YAMLElement) -> bool:
                            if _arg_4.tag == 0:
                                def predicate_2(x_1: str, _arg_4: Any=_arg_4) -> bool:
                                    return x_1 == _arg_4.fields[0].Value

                                return exists(predicate_2, field_names)

                            else: 
                                return False


                        overflow: FSharpList[tuple[str, YAMLElement]] = map_1(mapping_2, filter(predicate_3, ele.fields[0]))
                        dict_2: Any = dict([])
                        def action(tupled_arg: tuple[str, YAMLElement]) -> None:
                            add_to_dict(dict_2, tupled_arg[0], tupled_arg[1])

                        iterate(action, overflow)
                        return dict_2

                    else: 
                        raise Exception((to_text(printf("%s: %A"))("Expected an object")(ele) + "\\nParameter name: ") + "value")


            return ObjectExpr511()

        self.multiple_optional: IMultipleOptionalGetter = _arrow512()
        def _arrow514(__unit: None=None) -> IOverflowGetter:
            _this_3: Any = self
            class ObjectExpr513(IOverflowGetter):
                def FieldList(self, field_names_1: FSharpList[str]) -> Any:
                    if ele.tag == 3:
                        def mapping_3(_arg_7: YAMLElement) -> tuple[str, YAMLElement]:
                            if _arg_7.tag == 0:
                                return (_arg_7.fields[0].Value, _arg_7.fields[1])

                            else: 
                                raise Exception((to_text(printf("%s: %A"))("Expected a mapping")(ele) + "\\nParameter name: ") + "value")


                        def predicate_5(_arg_6: YAMLElement) -> bool:
                            if _arg_6.tag == 0:
                                def predicate_4(x_2: str, _arg_6: Any=_arg_6) -> bool:
                                    return x_2 == _arg_6.fields[0].Value

                                return not exists(predicate_4, field_names_1)

                            else: 
                                return False


                        overflow_1: FSharpList[tuple[str, YAMLElement]] = map_1(mapping_3, filter(predicate_5, ele.fields[0]))
                        dict_3: Any = dict([])
                        def action_1(tupled_arg_1: tuple[str, YAMLElement]) -> None:
                            add_to_dict(dict_3, tupled_arg_1[0], tupled_arg_1[1])

                        iterate(action_1, overflow_1)
                        return dict_3

                    else: 
                        raise Exception((to_text(printf("%s: %A"))("Expected an object")(ele) + "\\nParameter name: ") + "value")


            return ObjectExpr513()

        self.overflow: IOverflowGetter = _arrow514()

    @property
    def Required(self, __unit: None=None) -> IRequiredGetter:
        __: Getter = self
        return __.required

    @property
    def Optional(self, __unit: None=None) -> IOptionalGetter:
        __: Getter = self
        return __.optional

    @property
    def MultipleOptional(self, __unit: None=None) -> IMultipleOptionalGetter:
        __: Getter = self
        return __.multiple_optional

    @property
    def Overflow(self, __unit: None=None) -> IOverflowGetter:
        __: Getter = self
        return __.overflow


Getter_reflection = _expr515

def Getter__ctor_D9C929(ele: YAMLElement) -> Getter:
    return Getter(ele)


def object(getter: Callable[[IGetters], _A], value: YAMLElement) -> Any:
    return getter(Getter__ctor_D9C929(value))


def read(yaml: str) -> YAMLElement:
    return read_1(yaml)


__all__ = ["Helper_isString", "int_1", "float_1", "char", "bool_1", "map", "dict_1", "datetime", "datetime_offset", "option", "string", "tuple2", "tuple3", "tuple4", "tuple5", "tuple6", "tuple7", "tuple8", "list_1", "array", "seq", "resizearray", "values", "Getter_reflection", "object", "read"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.list import (of_array, FSharpList)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (map, iterate_indexed)
from ..fable_library.types import Array
from .data_context import (DataContext, DataContext__get_Explication, DataContext__set_Explication_279AAFF2, DataContext__get_Unit, DataContext__set_Unit_279AAFF2, DataContext__get_ObjectType, DataContext__set_ObjectType_279AAFF2, DataContext__get_Description, DataContext__set_Description_6DFDD678, DataContext__Copy)
from .Helper.hash_codes import (hash_1, box_hash_seq)
from .ontology_annotation import OntologyAnnotation
from .Table.composite_cell import CompositeCell
from .Table.composite_header import (IOType, CompositeHeader)

DataMapAux_dataHeader: CompositeHeader = CompositeHeader(11, IOType(2))

DataMapAux_explication: OntologyAnnotation = OntologyAnnotation("Clarification", "NCIT", "http://purl.obolibrary.org/obo/NCIT_C94778")

DataMapAux_explicationHeader: CompositeHeader = CompositeHeader(3, DataMapAux_explication)

DataMapAux_unit: OntologyAnnotation = OntologyAnnotation("Unit", "UO", "http://purl.obolibrary.org/obo/UO_0000000")

DataMapAux_unitHeader: CompositeHeader = CompositeHeader(3, DataMapAux_unit)

DataMapAux_objectType: OntologyAnnotation = OntologyAnnotation("Data Type", "NCIT", "http://purl.obolibrary.org/obo/NCIT_C42645")

DataMapAux_objectTypeHeader: CompositeHeader = CompositeHeader(3, DataMapAux_objectType)

DataMapAux_labelHeader: CompositeHeader = CompositeHeader(13, "Label")

DataMapAux_descriptionHeader: CompositeHeader = CompositeHeader(13, "Description")

DataMapAux_generatedByHeader: CompositeHeader = CompositeHeader(13, "Generated By")

DataMapAux_allowedHeaders: FSharpList[CompositeHeader] = of_array([DataMapAux_dataHeader, DataMapAux_explicationHeader, DataMapAux_unitHeader, DataMapAux_objectTypeHeader, DataMapAux_descriptionHeader, DataMapAux_generatedByHeader])

def DataMapAux_getOntologyColumn(f: Callable[[DataContext], OntologyAnnotation | None], data_contexts: Array[DataContext]) -> Array[CompositeCell]:
    def mapping(dc: DataContext, f: Any=f, data_contexts: Any=data_contexts) -> CompositeCell:
        match_value: OntologyAnnotation | None = f(dc)
        if match_value is None:
            return CompositeCell.empty_term()

        else: 
            return CompositeCell(0, match_value)


    return list(map(mapping, data_contexts))


def DataMapAux_getStringColumn(f: Callable[[DataContext], str | None], data_contexts: Array[DataContext]) -> Array[CompositeCell]:
    def mapping(dc: DataContext, f: Any=f, data_contexts: Any=data_contexts) -> CompositeCell:
        match_value: str | None = f(dc)
        if match_value is None:
            return CompositeCell.empty_free_text()

        else: 
            return CompositeCell(1, match_value)


    return list(map(mapping, data_contexts))


def DataMapAux_setOntologyColumn(f: Callable[[DataContext, OntologyAnnotation | None], None], column: Array[CompositeCell], data_contexts: Array[DataContext]) -> None:
    def action(i: int, cell: CompositeCell, f: Any=f, column: Any=column, data_contexts: Any=data_contexts) -> None:
        if cell.tag == 0:
            f(data_contexts[i], cell.fields[0])


    iterate_indexed(action, column)


def DataMapAux_setStringColumn(f: Callable[[DataContext, str | None], None], column: Array[CompositeCell], data_contexts: Array[DataContext]) -> None:
    def action(i: int, cell: CompositeCell, f: Any=f, column: Any=column, data_contexts: Any=data_contexts) -> None:
        if cell.tag == 1:
            f(data_contexts[i], cell.fields[0])


    iterate_indexed(action, column)


def DataMapAux_SanityChecks_rowIndexInBoundaries(row: int, data_contexts: Array[DataContext]) -> None:
    if row < 0:
        raise Exception("Row index must be greater or equal to 0.")

    if row >= len(data_contexts):
        raise Exception("Row index must be less than the number of rows.")



def DataMapAux_SanityChecks_lengthOfNewColumn(new_column: Array[CompositeCell], data_contexts: Array[DataContext]) -> None:
    if len(new_column) != len(data_contexts):
        raise Exception("Length of new column does not match length of data contexts.")



def _expr887() -> TypeInfo:
    return class_type("ARCtrl.DataMap", None, DataMap)


class DataMap:
    def __init__(self, data_contexts: Array[DataContext]) -> None:
        self.static_hash: int = 0
        self.data_contexts_0040108: Array[DataContext] = data_contexts

    @property
    def StaticHash(self, __unit: None=None) -> int:
        this: DataMap = self
        return this.static_hash

    @StaticHash.setter
    def StaticHash(self, value: int) -> None:
        this: DataMap = self
        this.static_hash = value or 0

    @property
    def DataContexts(self, __unit: None=None) -> Array[DataContext]:
        this: DataMap = self
        return this.data_contexts_0040108

    @staticmethod
    def init(__unit: None=None) -> DataMap:
        return DataMap([])

    def GetExplicationColumn(self, __unit: None=None) -> Array[CompositeCell]:
        this: DataMap = self
        return DataMapAux_getOntologyColumn(DataContext__get_Explication, this.data_contexts_0040108)

    def SetExplicationColumn(self, cells: Array[CompositeCell]) -> None:
        this: DataMap = self
        def _arrow880(dc: DataContext, oa: OntologyAnnotation | None=None) -> None:
            DataContext__set_Explication_279AAFF2(dc, oa)

        DataMapAux_setOntologyColumn(_arrow880, cells, this.data_contexts_0040108)

    def GetUnitColumn(self, __unit: None=None) -> Array[CompositeCell]:
        this: DataMap = self
        return DataMapAux_getOntologyColumn(DataContext__get_Unit, this.data_contexts_0040108)

    def SetUnitColumn(self, cells: Array[CompositeCell]) -> None:
        this: DataMap = self
        def _arrow881(dc: DataContext, oa: OntologyAnnotation | None=None) -> None:
            DataContext__set_Unit_279AAFF2(dc, oa)

        DataMapAux_setOntologyColumn(_arrow881, cells, this.data_contexts_0040108)

    def GetObjectTypeColumn(self, __unit: None=None) -> Array[CompositeCell]:
        this: DataMap = self
        return DataMapAux_getOntologyColumn(DataContext__get_ObjectType, this.data_contexts_0040108)

    def SetDataTypeColumn(self, cells: Array[CompositeCell]) -> None:
        this: DataMap = self
        def _arrow882(dc: DataContext, oa: OntologyAnnotation | None=None) -> None:
            DataContext__set_ObjectType_279AAFF2(dc, oa)

        DataMapAux_setOntologyColumn(_arrow882, cells, this.data_contexts_0040108)

    def GetLabelColumn(self, __unit: None=None) -> Array[CompositeCell]:
        this: DataMap = self
        def _arrow883(dc: DataContext) -> str | None:
            return dc.Name

        return DataMapAux_getStringColumn(_arrow883, this.data_contexts_0040108)

    def SetLabelColumn(self, cells: Array[CompositeCell]) -> None:
        this: DataMap = self
        def _arrow884(dc: DataContext, s: str | None=None) -> None:
            dc.Name = s

        DataMapAux_setStringColumn(_arrow884, cells, this.data_contexts_0040108)

    def GetDescriptionColumn(self, __unit: None=None) -> Array[CompositeCell]:
        this: DataMap = self
        return DataMapAux_getStringColumn(DataContext__get_Description, this.data_contexts_0040108)

    def SetDescriptionColumn(self, cells: Array[CompositeCell]) -> None:
        this: DataMap = self
        def _arrow885(dc: DataContext, s: str | None=None) -> None:
            DataContext__set_Description_6DFDD678(dc, s)

        DataMapAux_setStringColumn(_arrow885, cells, this.data_contexts_0040108)

    def GetDataContext(self, row: int, SkipValidation: Any | None=None) -> DataContext:
        this: DataMap = self
        DataMapAux_SanityChecks_rowIndexInBoundaries(row, this.data_contexts_0040108)
        return this.data_contexts_0040108[row]

    @staticmethod
    def get_data_context(row: int, SkipValidation: Any | None=None) -> Callable[[DataMap], DataContext]:
        def _arrow886(dm: DataMap) -> DataContext:
            return dm.GetDataContext(row, SkipValidation)

        return _arrow886

    def Copy(self, __unit: None=None) -> DataMap:
        this: DataMap = self
        return DataMap(list(map(DataContext__Copy, this.data_contexts_0040108)))

    def __eq__(self, obj: Any=None) -> bool:
        this: DataMap = self
        return hash_1(this) == hash_1(obj)

    def __hash__(self, __unit: None=None) -> Any:
        this: DataMap = self
        return box_hash_seq(this.data_contexts_0040108)

    @staticmethod
    def FileName() -> str:
        return "isa.datamap.xlsx"


DataMap_reflection = _expr887

def DataMap__ctor_4E3220A7(data_contexts: Array[DataContext]) -> DataMap:
    return DataMap(data_contexts)


__all__ = ["DataMapAux_dataHeader", "DataMapAux_explication", "DataMapAux_explicationHeader", "DataMapAux_unit", "DataMapAux_unitHeader", "DataMapAux_objectType", "DataMapAux_objectTypeHeader", "DataMapAux_labelHeader", "DataMapAux_descriptionHeader", "DataMapAux_generatedByHeader", "DataMapAux_allowedHeaders", "DataMapAux_getOntologyColumn", "DataMapAux_getStringColumn", "DataMapAux_setOntologyColumn", "DataMapAux_setStringColumn", "DataMapAux_SanityChecks_rowIndexInBoundaries", "DataMapAux_SanityChecks_lengthOfNewColumn", "DataMap_reflection"]




from __future__ import annotations
from ...arctrl_core.Helper.identifier import create_missing_identifier
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import LDNode
from .computational_workflow import LDComputationalWorkflow
from .file import LDFile
from .lab_protocol import LDLabProtocol
from .software_source_code import LDSoftwareSourceCode

def _expr1539() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDWorkflowProtocol", None, LDWorkflowProtocol)


class LDWorkflowProtocol:
    @staticmethod
    def schema_type() -> Array[str]:
        return [LDFile.schema_type(), LDComputationalWorkflow.schema_type(), LDSoftwareSourceCode.schema_type(), LDLabProtocol.schema_type()]

    @staticmethod
    def validate(wp: LDNode, context: LDContext | None=None) -> bool:
        return LDLabProtocol.validate(wp, context) if (LDSoftwareSourceCode.validate(wp, context) if LDComputationalWorkflow.validate(wp, context) else False) else False

    @staticmethod
    def create(id: str | None=None, inputs: Array[LDNode] | None=None, outputs: Array[LDNode] | None=None, creator: LDNode | None=None, date_created: str | None=None, licenses: Array[LDNode] | None=None, name: str | None=None, programming_languages: Array[LDNode] | None=None, sd_publisher: LDNode | None=None, url: str | None=None, version: str | None=None, description: str | None=None, has_parts: Array[LDNode] | None=None, intended_use: LDNode | None=None, comments: Array[LDNode] | None=None, computational_tools: Array[LDNode] | None=None, additional_type: Array[str] | None=None, context: LDContext | None=None) -> LDNode:
        at: Array[str] = ["WorkflowProtocol"]
        wp: LDNode = LDNode(clean(("#ComputationalWorkflow_" + create_missing_identifier()) + "") if (id is None) else id, LDWorkflowProtocol.schema_type(), at, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.input(), inputs, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.output(), outputs, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.creator(), creator, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.date_created(), date_created, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.license(), licenses, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.name(), name, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.programming_language(), programming_languages, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.sd_publisher(), sd_publisher, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.url(), url, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.version(), version, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.description(), description, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.has_part(), has_parts, context)
        wp.SetOptionalProperty(LDLabProtocol.intended_use(), intended_use, context)
        wp.SetOptionalProperty(LDLabProtocol.computational_tool(), computational_tools, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.additional_type(), additional_type, context)
        wp.SetOptionalProperty(LDComputationalWorkflow.comment(), comments, context)
        return wp


LDWorkflowProtocol_reflection = _expr1539

__all__ = ["LDWorkflowProtocol_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ..fable_library.list import (map, choose, of_array)
from ..fable_library.option import (default_arg, map as map_1)
from ..fable_library.reflection import (TypeInfo, class_type, int32_type, string_type, record_type)
from ..fable_library.reg_exp import (match, create)
from ..fable_library.string_ import (join, to_text, printf, replace)
from ..fable_library.system_text import (StringBuilder__ctor, StringBuilder__Append_Z721C83C5)
from ..fable_library.types import (to_string, Record)
from ..fable_library.util import (equals, ignore)
from .Helper.hash_codes import (box_hash_array, box_hash_option)
from .Helper.regex import Pattern_handleGroupPatterns

def _expr700() -> TypeInfo:
    return class_type("ARCtrl.Comment", None, Comment)


class Comment:
    def __init__(self, name: str | None=None, value: str | None=None) -> None:
        self._name: str | None = name
        self._value: str | None = value

    @property
    def Name(self, __unit: None=None) -> str | None:
        this: Comment = self
        return this._name

    @Name.setter
    def Name(self, name: str | None=None) -> None:
        this: Comment = self
        this._name = name

    @property
    def Value(self, __unit: None=None) -> str | None:
        this: Comment = self
        return this._value

    @Value.setter
    def Value(self, value: str | None=None) -> None:
        this: Comment = self
        this._value = value

    @staticmethod
    def make(name: str | None=None, value: str | None=None) -> Comment:
        return Comment(name, value)

    @staticmethod
    def create(name: str | None=None, value: str | None=None) -> Comment:
        return Comment.make(name, value)

    @staticmethod
    def to_string(comment: Comment) -> tuple[str, str]:
        return (default_arg(comment.Name, ""), default_arg(comment.Value, ""))

    def Copy(self, __unit: None=None) -> Comment:
        this: Comment = self
        name: str | None = this.Name
        value: str | None = this.Value
        return Comment.make(name, value)

    def __eq__(self, obj: Any=None) -> bool:
        this: Comment = self
        return (equals(obj.Value, this.Value) if equals(obj.Name, this.Name) else False) if isinstance(obj, Comment) else False

    def __hash__(self, __unit: None=None) -> Any:
        this: Comment = self
        return box_hash_array([box_hash_option(this.Name), box_hash_option(this.Value)])

    def __str__(self, __unit: None=None) -> str:
        this: Comment = self
        sb: Any = StringBuilder__ctor()
        ignore(StringBuilder__Append_Z721C83C5(sb, "Comment {"))
        def mapping_1(tupled_arg_1: tuple[str, str]) -> str:
            return to_text(printf("%s = \"%s\""))(tupled_arg_1[0])(tupled_arg_1[1])

        def chooser(tupled_arg: tuple[str, str | None]) -> tuple[str, str] | None:
            def mapping(o: str, tupled_arg: Any=tupled_arg) -> tuple[str, str]:
                return (tupled_arg[0], o)

            return map_1(mapping, tupled_arg[1])

        ignore(StringBuilder__Append_Z721C83C5(sb, join(", ", map(mapping_1, choose(chooser, of_array([("Name", this.Name), ("Value", this.Value)]))))))
        ignore(StringBuilder__Append_Z721C83C5(sb, "}"))
        return to_string(sb)

    @staticmethod
    def from_string(s: str) -> Comment:
        name_result: Any = match(create(Pattern_handleGroupPatterns("Name = \"[^\"]*\"")), s)
        value_result: Any = match(create(Pattern_handleGroupPatterns("Value = \"[^\"]*\"")), s)
        return Comment(replace(replace(name_result[0], "Name = ", ""), "\"", "") if (name_result is not None) else None, replace(replace(value_result[0], "Value = ", ""), "\"", "") if (value_result is not None) else None)


Comment_reflection = _expr700

def Comment__ctor_40457300(name: str | None=None, value: str | None=None) -> Comment:
    return Comment(name, value)


def _expr701() -> TypeInfo:
    return record_type("ARCtrl.Remark", [], Remark, lambda: [("Line", int32_type), ("Value", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class Remark(Record):
    Line: int
    Value: str
    @staticmethod
    def make(line: int, value: str) -> Remark:
        return Remark(line, value)

    @staticmethod
    def create(line: int, value: str) -> Remark:
        return Remark.make(line, value)

    @staticmethod
    def to_tuple(remark: Remark) -> tuple[int, str]:
        return (remark.Line, remark.Value)

    def Copy(self, __unit: None=None) -> Remark:
        this: Remark = self
        return Remark.make(this.Line, this.Value)


Remark_reflection = _expr701

__all__ = ["Comment_reflection", "Remark_reflection"]




from __future__ import annotations
from collections.abc import Callable
from json import JSONDecodeError
import numbers
import json as json_1
from typing import (Any, TypeVar)
from ..fable_library.result import FSharpResult_2
from ..fable_library.types import (Array, float32)
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (Advanced_fromValue, Helpers_prependPath, error_to_string)
from ..thoth_json_core.types import (IDecoderHelpers_1, Decoder_1, ErrorReason_1)

_T = TypeVar("_T")

class ObjectExpr386(IDecoderHelpers_1[Any]):
    def is_string(self, json_value: Any=None) -> bool:
        return str(type(json_value)) == "<class \'str\'>"

    def is_number(self, json_value_1: Any=None) -> bool:
        return (not (isinstance(json_value_1, (bool)))) if (isinstance(json_value_1, numbers.Number)) else False

    def is_boolean(self, json_value_2: Any=None) -> bool:
        return str(type(json_value_2)) == "<class \'bool\'>"

    def is_null_value(self, json_value_3: Any=None) -> bool:
        return json_value_3 is None

    def is_array(self, json_value_4: Any=None) -> bool:
        return isinstance(json_value_4, (list))

    def is_object(self, json_value_5: Any=None) -> bool:
        return isinstance(json_value_5, (dict))

    def has_property(self, field_name: str, json_value_6: Any=None) -> bool:
        return field_name in json_value_6

    def is_integral_value(self, json_value_7: Any=None) -> bool:
        return isinstance(json_value_7, (int))

    def as_string(self, json_value_8: Any=None) -> str:
        return json_value_8

    def as_boolean(self, json_value_9: Any=None) -> bool:
        return json_value_9

    def as_array(self, json_value_10: Any=None) -> Array[Any]:
        return json_value_10

    def as_float(self, json_value_11: Any=None) -> float:
        return json_value_11

    def as_float32(self, json_value_12: Any=None) -> float32:
        return json_value_12

    def as_int(self, json_value_13: Any=None) -> int:
        return json_value_13

    def get_properties(self, json_value_14: Any=None) -> IEnumerable_1[str]:
        return json_value_14.keys()

    def get_property(self, field_name_1: str, json_value_15: Any=None) -> Any:
        return json_value_15[field_name_1]

    def any_to_string(self, json_value_16: Any=None) -> str:
        return json_1.dumps(json_value_16, indent = 4)


Decode_helpers: IDecoderHelpers_1[Any] = ObjectExpr386()

def Decode_fromValue(decoder: Decoder_1[Any]) -> Callable[[Any], FSharpResult_2[_T, str]]:
    def _arrow387(value: Any=None, decoder: Any=decoder) -> FSharpResult_2[_T, str]:
        return Advanced_fromValue(Decode_helpers, decoder, value)

    return _arrow387


def Decode_fromString(decoder: Decoder_1[Any], value: str) -> FSharpResult_2[Any, str]:
    try: 
        json: Any = json_1.loads(value)
        match_value: FSharpResult_2[_T, tuple[str, ErrorReason_1[Any]]] = decoder.Decode(Decode_helpers, json)
        if match_value.tag == 1:
            final_error: tuple[str, ErrorReason_1[Any]]
            tupled_arg: tuple[str, ErrorReason_1[Any]] = match_value.fields[0]
            final_error = Helpers_prependPath("$", tupled_arg[0], tupled_arg[1])
            return FSharpResult_2(1, error_to_string(Decode_helpers, final_error[0], final_error[1]))

        else: 
            return FSharpResult_2(0, match_value.fields[0])


    except Exception as match_value_1:
        if isinstance(match_value_1, JSONDecodeError):
            return FSharpResult_2(1, "Given an invalid JSON: " + str(match_value_1))

        else: 
            raise match_value_1




def Decode_unsafeFromString(decoder: Decoder_1[Any], value: str) -> Any:
    match_value: FSharpResult_2[_T, str] = Decode_fromString(decoder, value)
    if match_value.tag == 1:
        raise Exception(match_value.fields[0])

    else: 
        return match_value.fields[0]



__all__ = ["Decode_helpers", "Decode_fromValue", "Decode_fromString", "Decode_unsafeFromString"]

